#!/usr/bin/env ruby

#############################################################################
#									    #
# Canny template library for Ruby					    #
# Copyright (C) 2004 Jason Wies <jason@xc.net>				    #
#									    #
# This library is free software; you can redistribute it and/or		    #
# modify it under the terms of the GNU Lesser General Public		    #
# License, version 2.1, as published by the Free Software Foundation.	    #
#									    #
# This library is distributed in the hope that it will be useful,	    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of	    #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	    #
# Lesser General Public License for more details.			    #
#									    #
# You should have received a copy of the GNU Lesser General Public	    #
# License along with this library; if not, write to the Free Software	    #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA #
#									    #
#############################################################################

require 'parsedate'
require 'digest/md5'

class Modifier
	@name = nil
	@function_name = nil

	attr_reader :name, :function_name

	def initialize(name, function_name)
		@name = name
		@function_name = function_name
	end
end

class Block
	@name = nil
	@parse_type = nil
	@contents = nil
	@subblocks = Array.new

	attr_reader :name, :parse_type, :subblocks
	protected :parse_type

	def initialize(contents)
		@name = self.class.name
		@parse_type = self.class.parse_type

		@contents = contents

		if @parse_type == 'block'
			@subblocks = Block.get_subblocks(self.strip_block)
		end
	end

	def Block.name
		@name
	end

	def Block.parse_type
		@parse_type
	end

	def get_params
		params = Hash.new

		if @parse_type == 'single' or @parse_type == 'block'
			index = @contents.index("}")
			str = @contents[0, index]
			str.gsub(/[[:space:]]([[:alnum:]_]+)="?(.*?)"?(?=([[:space:]]|\}))/) do |match|
				params[$1] = $2
			end
		end

		params
	end

	def generate_code
		str = ''
		str += self.generate_code_subblocks
		str
	end

	def generate_code_subblocks
		str = ''
		if @subblocks
			@subblocks.each { |block| str += block.generate_code }
		end
		str
	end

	def replace_vars(str, no_braces=false, do_to_s=true)
		str = str.dup
		str.gsub!(/\.(\$[[:alnum:]_]+)/, '[\1]')
		str.gsub!(/\.([[:alnum:]_]+)/, '[\'\1\']')
		str.gsub!(/\$([[:alnum:]_]+)/, 'vars[\'\1\']')

		str.gsub!(/\[(vars\[([[:alnum:]_'"\[\]])+\])\]/) do |match|
			if do_to_s
				match = '[' + $1 + '.to_s' + ']'
			else
				match
			end
		end

		str.gsub!(/vars\[([[:alnum:]_'"\[\]\.])+\]/) do |match|
			if do_to_s
				match += '.to_s'
			end

			if not no_braces
				match = '#{' + match + '}'
			else
				match
			end
		end

		str
	end

	def print_names
		puts @name

		if @subblocks
			@subblocks.each { |block| block.print_names }
		end
	end

	def print_tree
		puts "\n#{@name}"

		@contents.each_line do |line|
			print "\t" + line
		end

		if @subblocks
			@subblocks.each { |block| block.print_tree }
		end
	end

	def find_end
		Block.find_end_generic(@name, @contents)
	end

	def Block.find_end_generic(name, str)
		count = 0
		found = false
		chars_read = 0
		index = 0
		parts = str.split("{")
		parts.each do |line|
			line = "{" + line
			index = 0
			while true
				index = line.index(/\{[[:space:]]*#{name}[[:space:]](.*?)\}.*/, index)

				if not index
					break
				else
					index += 1
				end

				count += 1
			end

			index = 0
			while true
				index = line.index(/\{[[:space:]]*\/#{name}[[:space:]]*\}/, index)

				if not index
					break
				else
					index += 1
				end

				count -= 1

				if count == 0
					found = true
					break
				end
			end

			if found
				index += chars_read
				break
			end

			chars_read += line.length
		end

		if found
			index = str.index("}", index)
			index
		else
			nil
		end
	end

	def strip_block()
		if @parse_type == 'single'
			index = @contents.index("}")
			str = @contents[index+1, @contents.length]
		elsif @parse_type == 'block'
			index = @contents.index("}")
			str = @contents[index+1, @contents.length]

			index = str.rindex("{")
			str = str[0, index]
		elsif @parse_type == 'open'
			tag_length = @name.length + 1;
			str = @contents[tag_length + 1, @contents.length - tag_length - 1]
		elsif @parse_type == 'text'
			str = ''
		end

		str
	end

	def Block.get_subblocks(contents)
		blocks = Array.new

		while contents and contents.length
			index = contents.index("{")
			if not index
				blocks << BlockText.new(contents)
				break
			end

			if index > 0
				blocks << BlockText.new(contents[0, index])
				contents = contents[index..contents.length]

				next
			end

			index_close = contents.index("}")
			index_var = contents.index(/^\{[[:space:]]*\$([[:alnum:]_]+)[[:space:]\[\|\.\$].*\}.*/)

			if index_var and index_close and index_var < index_close
				blocks << BlockVariable.new(contents[0..index_close])
				contents = contents[index_close+1..contents.length]

				next
			end

			name = contents.sub(/\{[[:space:]]*([[:alpha:]_]+|\*)(.*?)\}.*/m, '\1')

			block = Canny.registered_blocks.detect { |block| block.name == name }
			if block
				if block.parse_type == 'single'
					index = contents.index("}")

					if not index
						puts "Fatal Canny Error: Missing } for \"#{name}\""
						exit
					end

					blocks << block.new(contents[0..index])
				elsif block.parse_type == 'block'
					index = Block.find_end_generic(block.name, contents)

					if not index
						puts "Fatal Canny Error: Missing end tag for \"#{name}\""
						exit
					end

					blocks << block.new(contents[0..index])
				elsif block.parse_type == 'open'
					index = contents.index("#{block.name}}")

					if not index
						puts "Fatal Canny Error: Missing end tag for \"#{name}\""
						exit
					end

					index = contents.index("}", index)
					blocks << block.new(contents[0..index])
				end
			else
				puts "Fatal Canny Error: Unknown block type \"#{name}\" (need to register it?)"
				exit
			end

			contents = contents[index+1..contents.length]
		end

		if blocks.length > 0
			return blocks
		else
			return nil
		end
	end
end

class Canny
	@vars = nil
	@empty_string_is_nil = false
	@zero_is_nil = false
	@@registered_blocks = Array.new
	@@registered_modifiers = Array.new

	attr_writer :empty_string_is_nil, :zero_is_nil

	def initialize
		@vars = Hash.new
		@empty_string_is_nil = false
		@zero_is_nil = false
		@@registered_blocks = self.class.registered_blocks
		@@registered_modifiers = self.class.registered_modifiers
	end

	def assign(var, value)
		@vars[var] = value
	end

	def template_dir=(val)
		if not val =~ /\/$/
			val += "/"
		end

		@template_dir = val
	end

	def compile_dir=(val)
		if not val =~ /\/$/
			val += "/"
		end

		@compile_dir = val
	end

	def Canny.registered_blocks
		@@registered_blocks
	end

	def Canny.registered_modifiers
		@@registered_modifiers
	end

	def Canny.register(block)
		@@registered_blocks << block
	end

	def Canny.register_modifier(name, function_name)
		modifier = Modifier.new(name, function_name)
		@@registered_modifiers << modifier
	end

	def fatal_error(err)
		puts "Fatal Canny Error:" + " <br />"
		puts err.message + " <br /><br />\n\n"
		puts err.backtrace.join(" <br />\n")
		exit
	end

	def fatal_error_str(str)
		puts "Fatal Canny Error:" + " <br />"
		puts str + " <br />\n"
		exit
	end

	def include_r(str)
		str.gsub!(/(\{[[:space:]]*include(?=[[:space:]]+)[[:space:]]file="(.*?)".*\})/) do |block|
			file = $2

			f = File.new(@template_dir + file, "r")
			include_contents = f.read
			f.close

			include_contents = include_r(include_contents)

			block + "\n" + include_contents + "\n" + "{/include}"
		end

		return str
	end

	def display(filename)
		begin
			if not @template_dir
				fatal_error_str("Please set the template_dir variable.")
			elsif not @compile_dir
				fatal_error_str("Please set the compile_dir variable.")
			end

			template_filename = @template_dir + filename

			file_hash = Digest::MD5.hexdigest(template_filename)
			compile_filename = @compile_dir + file_hash + ".ctpl"
			compile_files_filename = @compile_dir + file_hash + "_files.ctpl"

			if not File.exists?(template_filename)
				fatal_error_str("#{template_filename} does not exist.")
			elsif not File.readable?(template_filename)
				fatal_error_str("#{template_filename} exists, but is not readable.")
			elsif not File.writable?(@compile_dir)
				fatal_error_str("Can't write compiled output.  Check permissions on #{@compile_dir}.")
			end

			parsed = ''

			if File.exists?(compile_filename) and File.exists?(compile_files_filename)
				compile_stat = File.stat(compile_filename)

				f = File.open(compile_files_filename)
				files_dependent = Marshal.load(f.read)
				f.close

				files_dependent << template_filename

				use_compiled = true
				files_dependent.each do |filename|
					filename_stat = File.stat(filename)

					if filename_stat.mtime >= compile_stat.mtime
						use_compiled = false
						break
					end
				end

				if use_compiled
					f = File.new(compile_filename)
					parsed = f.read
					f.close
				end
			end

			if parsed.length == 0
				f = File.new(template_filename, "r")
				contents = f.read
				f.close

				contents = include_r(contents)

				files_dependent = Array.new
				contents.scan(/(\{[[:space:]]*include(?=[[:space:]]+)[[:space:]]file="(.*?)".*\})/) do |block|
					files_dependent << @template_dir + $2
				end

				contents = "{all}" + contents + "{/all}"

				parsed += <<-EOS
					class ReallyNilClass
						def nil?
							return true
						end

						def to_s
							return nil
						end
					end

					class ReallyFalseClass
						def to_s
							return false
						end
					end

					class CannyClass
						@hashes = Hash.new

						def initialize
							@hashes = Hash.new
							@hashes['foreach'] = Hash.new
							@hashes['section'] = Hash.new
							@hashes['capture'] = Hash.new
							@hashes['template'] = ''
						end

						def [](key)
							if key == 'now'
								self.now
							else
								@hashes[key]
							end
						end

						def []=(key, value)
							@hashes[key] = value
						end

						def now
							Time.now.to_i
						end
					end

					class CannyProcClass < Proc
						def to_s
							self.call
						end
					end

					vars['canny'] = CannyClass.new
					vars['canny']['template'] = '#{filename}'

					def nil_r(var)
						if var.kind_of?(Hash)
							var.each_key do |key|
								if var[key] == nil or (var[key] == '' and $settings['empty_string_is_nil']) or (var[key] == 0 and $settings['zero_is_nil'])
									var[key] = ReallyNilClass.new
								elsif var[key] == false
									var[key] = ReallyFalseClass.new
								end

								nil_r(var[key])
							end
							var.default = ReallyNilClass.new
						end
					end

					nil_r(vars)

					def array_to_hash(var)
						var_new = Hash.new

						var.each_index do |key|
							if key.kind_of?(Numeric)
								key_new = key.to_s
							else
								key_new = key
							end

							var_new[key_new] = var[key]
						end

						return var_new
					end

					def to_string_r(vars)
						if vars.kind_of?(Hash) or vars.kind_of?(Array)
							if vars.kind_of?(Hash)
								keys = vars.keys
							else
								keys = Array.new
								vars.each_index { |key| keys << key }
							end

							keys.each do |key|
								var = vars[key]

								if var.kind_of?(Array)
									vars[key] = to_string_r(vars[key])
									vars[key] = array_to_hash(var)
								elsif var.kind_of?(Hash)
									vars[key] = to_string_r(vars[key])
								elsif var.kind_of?(Numeric)
									vars[key] = var.to_s
								else
									vars[key] = var
								end
							end
						end

						return vars
					end

					vars = to_string_r(vars.dup)
				EOS

				all = BlockAll.new(contents)
				parsed += all.generate_code

				f = File.open(compile_filename, 'w', 0600)
				f.write(parsed)
				f.close

				files_dependent_str = Marshal.dump(files_dependent)

				f = File.open(compile_files_filename, 'w', 0600)
				f.write(files_dependent_str)
				f.close
			end

			vars_header = ''
			vars_header += "vars_string = '" + Marshal.dump(@vars).sub("'", "\'") + "'" + "\n"
			vars_header += "vars = Marshal.load(vars_string)" + "\n"

			settings_header = <<-EOS
				$settings = Hash.new
				$settings['empty_string_is_nil'] = #{@empty_string_is_nil} ? true : false
				$settings['zero_is_nil'] = #{@zero_is_nil} ? true : false
			EOS

			parsed = vars_header + "\n" + settings_header + "\n" + parsed

			eval(parsed)
#			print parsed
		rescue => err
			self.fatal_error(err)
		end
	end
end

class BlockAll < Block
	@name = 'all'
	@parse_type = 'block'
end

class BlockIf < Block
	@name = 'if'
	@parse_type = 'block'

	def generate_code
		str = ''
		str += @contents.sub(/\{[[:space:]]*#{@name}[[:space:]]*(.*?)\}.*/m, 'if \1' + "\n")
		str = self.replace_vars(str, true) + "\n"
		str += self.generate_code_subblocks
		str += "end\n"
		str
	end
end

class BlockElseIf < Block
	@name = 'elseif'
	@parse_type = 'single'

	def generate_code
		str = ''
		str += @contents.sub(/\{[[:space:]]*#{@name}[[:space:]]*(.*?)\}/, 'elsif \1' + "\n")
		str = self.replace_vars(str, true)
		subblocks_str = self.generate_code_subblocks
		if subblocks_str
			subblocks_str.gsub!(/(^|\n)/, "\t")
			str += subblocks_str
		end
		str
	end
end

class BlockElse < Block
	@name = 'else'
	@parse_type = 'single'

	def generate_code
		str = ''
		str += @contents.sub(/\{[[:space:]]*#{@name}[[:space:]]*(.*?)\}/, 'else \1' + "\n")
		str = self.replace_vars(str, true)
		subblocks_str = self.generate_code_subblocks
		if subblocks_str
			subblocks_str.gsub!(/(^|\n)/, "\t")
			str += subblocks_str
		end
		str
	end
end

class BlockVariable < Block
	@name = 'variable'
	@parse_type = 'single'

	def generate_code
		str = @contents.dup
		str.sub!(/\{[[:blank:]]*(.*?)[[:blank:]]*\}/, '\1')
		parts = str.split("|")
		str = parts[0]

		str = self.replace_vars(str)
		str = '"' + str + '"'

		parts[1..parts.size].each do |part|
			args = part.split(":")

			name = args[0]
			args = args[1..args.size]

			modifier = Canny.registered_modifiers.detect { |modifier| modifier.name == name }

			if modifier
				if args.size > 0
					args_str = ", ["
					args.each do |arg|
						if arg =~ /^'(.*)'$/
							arg.sub!(/^'/, '')
							arg.sub!(/'$/, '')
							arg2 = arg.dup
						else
							arg.sub!(/^"/, '')
							arg.sub!(/"$/, '')
							arg2 = self.replace_vars(arg, true)
						end

						if arg != arg2
							space_before = arg2.sub(/^([[:space:]]*).*/, '\1')
							space_after = arg2.sub(/.*([[:space:]]*)$/, '\1')

							args_str += "'#{space_before}' + " + "#{arg2}" + " + '#{space_after}', "
						else
							args_str += "'#{arg}', "
						end
					end
					args_str += "]"
				else
					args_str = ""
				end

				str = modifier.function_name + "(" + str + args_str + ")"
			else
				puts "Fatal Canny Error: Unknown modifier #{name} specified (need to register it?)"
				exit
			end
		end

		str = "print " + str + "\n"

		str
	end
end

class BlockText < Block
	@name = 'text'
	@parse_type = 'text'

	def generate_code
		if @contents == "\n"
			"print \"\\n\"\n"
		else
			contents_str = @contents.gsub('"', '\"')
			"print \"#{contents_str}\"\n"
		end
	end
end

class BlockForeach < Block
	@name = 'foreach'
	@parse_type = 'block'

	def generate_code
		str = ''

		params = self.get_params

		if params['from'] and params['from'].length > 0
			from = self.replace_vars(params['from'], true, false)
		else
			puts "Fatal Canny Error: No \"from\" specified in foreach"
			exit
		end

		if params['key'] and params['key'].length > 0
			key = params['key']
		else
			key = 'key'
		end

		if params['item'] and params['item'].length > 0
			item = params['item']
		else
			item = 'item'
		end

		if params['name'] and params['name'].length > 0
			name = params['name']
		else
			name = 'this'
		end

		if params['sort'] and params['sort'].length > 0
			sort = self.replace_vars(params['sort'], false, false)
		else
			sort = 'keys'
		end

		if params['show'] and params['show'].length > 0
			show = self.replace_vars(params['show'], true, false)
		else
			show = true
		end

		str += <<-EOS
			key_bak = vars['#{key}']
			item_bak = vars['#{item}']

			vars['canny']['foreach']['#{name}'] = Hash.new
			vars['canny']['foreach']['#{name}']['iteration'] = 0
			vars['canny']['foreach']['#{name}']['first'] = true
			vars['canny']['foreach']['#{name}']['last'] = #{from}.size <= 1 ? true : ReallyFalseClass.new
			vars['canny']['foreach']['#{name}']['total'] = #{from}.size

			if "#{sort}".to_s =~ /^values$/i
				sort = 'values'
			elsif "#{sort}".to_s =~ /^none$/i
				sort = 'none'
			else
				sort = 'keys'
			end

			if #{from} and (#{from}.kind_of?(Array) or #{from}.kind_of?(Hash)) and #{from}.size > 0 and not (#{show}.nil? or #{show}.to_s == false or #{show}.to_s == 'false')
				if sort == 'keys'
					if #{from}.kind_of?(Array)
						keys_sorted = Array.new
						#{from}.each_index { |key| keys_sorted << key }
						keys_sorted.sort!
					elsif #{from}.kind_of?(Hash)
						keys_sorted = #{from}.keys.sort
					end
				elsif sort == 'values'
					if #{from}.kind_of?(Array)
						keys_sorted = #{from}.sort.collect {|x| #{from}.index(x) }
					elsif #{from}.kind_of?(Hash)
						keys_sorted = #{from}.sort {|x,y| x[1] <=> y[1]}
						keys_sorted.collect! {|x,y| x}
					end
				elsif sort == 'none'
					if #{from}.kind_of?(Array)
						keys_sorted = Array.new
						#{from}.each_index { |key| keys_sorted << key }
					elsif #{from}.kind_of?(Hash)
						keys_sorted = #{from}.keys
					end
				end

				keys_sorted.each do |vars['#{key}']|
					if vars['canny']['foreach']['#{name}']['iteration'] > 0
						vars['canny']['foreach']['#{name}']['first'] = ReallyFalseClass.new
						vars['canny']['foreach']['#{name}']['last'] = vars['canny']['foreach']['#{name}']['iteration'] == #{from}.size - 1 ? true : ReallyFalseClass.new
					end

					vars['canny']['foreach']['#{name}']['iteration'] += 1
					vars['#{item}'] = #{from}[vars['#{key}']]
		EOS

		subblocks_str = self.generate_code_subblocks

		if subblocks_str
			subblocks_str.gsub!(/(^|\n)/, '\1' + "\t")
			str += subblocks_str
		end

		if not subblocks.detect { |block| block.name == 'foreachelse' }
			str += "end" + "\n"
		end

		str += <<-EOS
			end
			vars['#{key}'] = key_bak
			vars['#{item}'] = item_bak
		EOS

		return str
	end
end

class BlockForeachElse < Block
	@name = 'foreachelse'
	@parse_type = 'single'

	def generate_code
		str = "\nend\nelse\n"
		str
	end
end

class BlockSection < Block
	@name = 'section'
	@parse_type = 'block'

	def generate_code
		str = ''

		params = self.get_params

		if params['name'] and params['name'].length > 0
			name = params['name']
		else
			name = 'index'
		end

		if params['loop'] and params['loop'].length > 0
			loop_var = self.replace_vars(params['loop'], true, false)
		else
			return ''
		end

		if params['start'] and params['start'].length > 0
			start = self.replace_vars(params['start'], true, false)
		else
			start = 0
		end

		if params['step'] and params['step'].length > 0
			step = self.replace_vars(params['step'], true, false)
		else
			step = 1
		end

		if params['max'] and params['max'].length > 0
			max = self.replace_vars(params['max'], true, false)
		else
			max = false
		end

		if params['show'] and params['show'].length > 0
			show = self.replace_vars(params['show'], true, false)
		else
			show = true
		end

		str += <<-EOS
			if (#{loop_var}.kind_of?(Array) or #{loop_var}.kind_of?(Hash)) and #{loop_var}.size > 0 and not (#{show}.nil? or #{show}.to_s == false or #{show}.to_s == 'false')
				if not #{start}.nil?
					start = #{start}
				else
					start = 0
				end

				if not #{step}.nil?
					step = #{step}
				else
					step = 1
				end

				if not #{max}.nil?
					max = #{max}
				else
					max = nil
				end

				name_bak = vars['#{name}']

				if #{loop_var}.kind_of?(Array)
					keys = Array.new
					#{loop_var}.each_index { |key| keys << key }
				else
					keys = #{loop_var}.keys
				end

				count = #{loop_var}.size
				highest_index_loop = start.to_i + (count * step.to_i)
				last_index = count < highest_index_loop ? count : highest_index_loop

				vars['canny']['section']['#{name}'] = Hash.new
				vars['canny']['section']['#{name}']['index'] = start.to_i
				vars['canny']['section']['#{name}']['index_prev'] = -1
				vars['canny']['section']['#{name}']['index_next'] = start.to_i + step.to_i
				vars['canny']['section']['#{name}']['iteration'] = 0
				vars['canny']['section']['#{name}']['rownum'] = 0
				vars['canny']['section']['#{name}']['loop'] = 0
				vars['canny']['section']['#{name}']['first'] = true
				vars['canny']['section']['#{name}']['last'] = start.to_i + step.to_i > last_index ? true : ReallyFalseClass.new
				vars['canny']['section']['#{name}']['total'] = ((count - start.to_i).to_f / step.to_i).ceil

				nr_times = vars['canny']['section']['#{name}']['total']

				vars['#{name}'] = start.to_i
				nr_times.times do
					if vars['canny']['section']['#{name}']['iteration'] != 0
						vars['#{name}'] += step.to_i
						vars['canny']['section']['#{name}']['index_prev'] = vars['canny']['section']['#{name}']['index']
						vars['canny']['section']['#{name}']['index'] += step.to_i
						vars['canny']['section']['#{name}']['index_next'] += step.to_i
						vars['canny']['section']['#{name}']['first'] = ReallyFalseClass.new
						vars['canny']['section']['#{name}']['last'] = vars['canny']['section']['#{name}']['index_next'] >= count ? true : ReallyFalseClass.new
					end

					if max and vars['canny']['section']['#{name}']['iteration'] >= max
						break
					end

					vars['canny']['section']['#{name}']['iteration'] += 1
					vars['canny']['section']['#{name}']['rownum'] += 1
					vars['canny']['section']['#{name}']['loop'] += 1

		EOS

		subblocks_str = self.generate_code_subblocks
		str += subblocks_str

		str += "end" + "\n";

		if not subblocks.detect { |block| block.name == 'sectionelse' }
			str += "end" + "\n"
		end

		str += "vars['#{name}'] = name_bak" + "\n"

		str
	end
end

class BlockSectionElse < Block
	@name = 'sectionelse'
	@parse_type = 'single'

	def generate_code
		str = "\nend\nelse\n"
		str
	end
end

class BlockComment < Block
	@name = '*'
	@parse_type = 'open'

	def generate_code
		return ''
	end
end

class BlockInclude < Block
	@name = 'include'
	@parse_type = 'block'

	def generate_code
		str = ''

		params = self.get_params

		str += "template_bak = vars['canny']['template']" + "\n"
		str += "vars['canny']['template'] = '" + params['file'] + "'" + "\n"

		params.each_key do |key|
			if key == "file"
				next
			end

			value = self.replace_vars(params[key].dup, true, false)
			if value != params[key]
				str += "vars['#{key}'] = #{value}" + "\n"
			else
				str += "vars['#{key}'] = \"#{value}\"" + "\n"
			end
		end

		str += self.generate_code_subblocks

		str += "vars['canny']['template'] = template_bak" + "\n"

		return str
	end
end

class BlockCapture < Block
	@name = 'capture'
	@parse_type = 'block'
	@@output_nr = 0

	def generate_code
		str = ''

		params = self.get_params

		if params['name'] and params['name'].length > 0
			name = params['name']
		else
			name = 'default'
		end

		if params['assign'] and params['assign'].length > 0
			assign = params['assign']
		else
			assign = nil
		end

		subblocks_str = self.generate_code_subblocks
		subblocks_str.gsub!("'", "\\\\'")

		var_name = "output_" + @@output_nr.to_s

		@@output_nr += 1

		str += "#{var_name} = '#{subblocks_str}'" + "\n"
		str += "vars['canny']['capture']['#{name}'] = CannyProcClass.new { eval(#{var_name}) }" + "\n"

		if assign
			str += "vars['#{assign}'] = CannyProcClass.new { eval(#{var_name}) }" + "\n"
		end

		str
	end
end

Canny.register(BlockText)
Canny.register(BlockAll)
Canny.register(BlockIf)
Canny.register(BlockElseIf)
Canny.register(BlockElse)
Canny.register(BlockVariable)
Canny.register(BlockForeach)
Canny.register(BlockForeachElse)
Canny.register(BlockSection)
Canny.register(BlockSectionElse)
Canny.register(BlockComment)
Canny.register(BlockInclude)
Canny.register(BlockCapture)

def modifier_canny_capitalize(str)
	str.split(" ").collect { |x| x.capitalize }.join(" ")
end

def modifier_canny_count_characters(str, params=[])
	if params[0] == 'true'
		include_whitespace = true
	else
		include_whitespace = false
	end

	if include_whitespace
		str.length
	else
		str.gsub(/[[:space:]]/, '').length
	end
end

def modifier_canny_cat(str, params=[])
	str += params[0]
end

def modifier_canny_count_paragraphs(str)
	if str.length == 0
		return 0
	end

	count = 1

	str.scan(/\n\n/) { count += 1 }

	count
end

def modifier_canny_count_sentences(str)
	if str.length == 0
		return 0
	end

	count = 0

	str.scan(/\.|\!|\?/) { count += 1 }

	count
end

def modifier_canny_count_words(str)
	str.split(" ").size
end

def modifier_canny_date_format(str, params=[])
	if params[0] and params[0].length > 0
		format_str = params[0]
	else
		format_str = "%b %d, %Y"
	end

	if not str or str.length == 0
		if params[1] and params[1].length > 0
			str = params[1]
		else
			return ""
		end
	end

	if str.to_i.to_s == str
		t = Time.at(str.to_i)
		t.strftime(format_str)
	else
		pd = ParseDate.parsedate(str)

		if pd[0]
			t = Time.utc(pd[4], pd[5], pd[3], pd[2], pd[1], pd[0], pd[7], nil, nil, pd[6])
			t.strftime(format_str)
		else
			''
		end
	end
end

def modifier_canny_default(str, params=[])
	if str and str.length > 0
		str
	elsif params[0] and params[0].length > 0
		params[0]
	else
		''
	end
end

def modifier_canny_indent(str, params=[])
	if params[0] and params[0].to_i.to_s == params[0]
		times = params[0].to_i
	else
		times = 4
	end

	if params[1]
		char = params[1]
		if char == '\t'
			char = "\t"
		end
	else
		char = " "
	end

	replace_with = char * times

	str = replace_with + str

	if str =~ /\n/
		str.gsub!(/(\n)/, '\1' + replace_with)
	end

	str
end

def modifier_canny_lower(str)
	str.downcase
end

def modifier_canny_nl2br(str)
	str.gsub("\n", "<br />")
end

def modifier_canny_regex_replace(str, params=[])
	if (not params[0] or params[0].length == 0) or (not params[1])
		return str
	end

	pattern = params[0]
	replace_with = params[1]

	index = pattern.rindex("/")
	if not index
		return str
	end

	options = pattern[index..str.length]

	pattern = pattern.sub(/\/(.*)\/([igm]*)$/, '\1')

	options_mask = 0

	if options.index("i")
		options_mask |= Regexp::IGNORECASE
	end

	if options.index("m")
		options_mask |= Regexp::MULTILINE
	end

	if options.index("e")
		options_mask |= Regexp::EXTENDED
	end

	regex = Regexp.new(pattern, options)

	if options.index("g")
		str.gsub(regex, replace_with)
	else
		str.sub(regex, replace_with)
	end
end

def modifier_canny_replace(str, params=[])
	if (not params[0] or params[0].length == 0) or not params[1]
		return str
	end

	str.gsub(params[0], params[1])
end

def modifier_canny_spacify(str, params=[])
	if params[0] and params[0].length > 0
		char = params[0]
	else
		char = " "
	end

	str.split("").join(char)
end

def modifier_canny_string_format(str, params=[])
	sprintf(params[0], str)
end

def modifier_canny_strip(str, params=[])
	if params[0] and params[0].length > 0
		char = params[0]
	else
		char = " "
	end

	str.gsub(/[[:space:]]+/, char)
end

def modifier_canny_strip_tags(str, params=[])
	if params[0] and params[0].length > 0
		if params[0] == 'false'
			char = ''
		else
			char = " "
		end
	else
		char = " "
	end

	str.gsub(/<(.*?)>/, char)
end

def modifier_canny_truncate(str, params=[])
	if params[0] and params[0].to_i.to_s == params[0]
		length = params[0].to_i
	else
		length = 80
	end

	if params[1]
		append = params[1]
	else
		append = '...'
	end

	if params[2]
		if params[2] == 'true'
			exact_character = true
		else
			exact_character = false
		end
	else
		exact_character = false
	end

	if str.length < (length + append.length)
		return str + append
	end

	str = str[0..(length - append.length)]
	index = str.rindex(" ")

	if exact_character or not index
		str + append
	else
		str[0..index-1] + append
	end
end

def modifier_canny_upper(str)
	str.upcase
end

def modifier_canny_wordwrap(str, params=[])
	if params[0] and params[0].to_i.to_s == params[0]
		length = params[0].to_i
	else
		length = 80
	end

	if params[1]
		char = params[1]

		if char == '\t'
			char = "\t"
		elsif char == '\n'
			char = "\n"
		end
	else
		char = "\n"
	end

	if params[2]
		if params[2] == 'true'
			exact_character = true
		else
			exact_character = false
		end
	else
		exact_character = false
	end

	str.gsub!(/(.*?)(\n|$)/) do |match|
		line = match.sub(/\n$/, '')

		if line.length <= length
			return line + char
		end

		if exact_character
			return line.gsub(/(.{1,#{length}})/, '\1' + char)
		else
			#line.gsub(/(.{1,#{length}})/, '')
		end
	end

	str
end

Canny.register_modifier('capitalize', 'modifier_canny_capitalize')
Canny.register_modifier('count_characters', 'modifier_canny_count_characters')
Canny.register_modifier('cat', 'modifier_canny_cat')
Canny.register_modifier('count_paragraphs', 'modifier_canny_count_paragraphs')
Canny.register_modifier('count_sentences', 'modifier_canny_count_sentences')
Canny.register_modifier('count_words', 'modifier_canny_count_words')
Canny.register_modifier('date_format', 'modifier_canny_date_format')
Canny.register_modifier('default', 'modifier_canny_default')
Canny.register_modifier('indent', 'modifier_canny_indent')
Canny.register_modifier('lower', 'modifier_canny_lower')
Canny.register_modifier('nl2br', 'modifier_canny_nl2br')
Canny.register_modifier('regex_replace', 'modifier_canny_regex_replace')
Canny.register_modifier('replace', 'modifier_canny_replace')
Canny.register_modifier('spacify', 'modifier_canny_spacify')
Canny.register_modifier('string_format', 'modifier_canny_string_format')
Canny.register_modifier('strip', 'modifier_canny_strip')
Canny.register_modifier('strip_tags', 'modifier_canny_strip_tags')
Canny.register_modifier('truncate', 'modifier_canny_truncate')
Canny.register_modifier('upper', 'modifier_canny_upper')
#Canny.register_modifier('wordwrap', 'modifier_canny_wordwrap')
